import typing
import logging


default_logger = logging.getLogger()


Logger = typing.Union[logging.Logger, logging.LoggerAdapter]


class ExecutorsPool:

    def __init__(
        self,
        executors: typing.Optional[list] = None,
        logger: Logger = default_logger,
    ):
        self._executors = executors or list()
        self.logger = logger

    def add(
        self,
        type_: typing.Type,
        executor: typing.Callable,
        *args,
        **kwargs,
    ) -> None:
        self._executors.append({
            'type': type_,
            'executor': executor,
            'args': args or list(),
            'kwargs': kwargs or dict(),
        })

    def remove(self, type_: typing.Type) -> None:
        self._executors = [
            executor
            for executor in self._executors
            if executor['type'] != type_
        ]

    def execute(self, data: typing.Any) -> typing.Any:
        for executor_data in self._executors:
            if isinstance(data, executor_data['type']):
                result = executor_data['executor'](
                    data,
                    *executor_data['args'],
                    **executor_data['kwargs'],
                )
                self.logger.debug(
                    f'Executor for {type(data)} executed', 
                    extra=dict(
                        data=data,
                        type=executor_data['type'],
                        executor=executor_data['type'].__name__,
                        func_args=[str(arg) for arg in executor_data['args']],
                        func_kwargs={
                            k: str(v) 
                            for k, v in executor_data['kwargs'].items()
                        },
                        result=result,
                    )
                )
                return result

        self.logger.debug(f'No executors for {type(data)}. Return raw')
        return data



import functools
import logging
import inspect
import typing
import loguru

from acenus.executors_pool import ExecutorsPool


default_logger = loguru.logger


Logger = typing.Union[logging.Logger, logging.LoggerAdapter]


def get_func_kwargs(func: typing.Callable, kwargs: dict) -> dict:
    """ Inspect and get function kwargs """
    parameters = tuple(inspect.signature(func).parameters.keys())
    if "kwargs" not in parameters:
        kwargs = {k: v for k, v in kwargs.items() if k in parameters}
    return kwargs


def log_function(func: typing.Callable, logger: Logger):
    """ Decorator for log functions with exceptions """

    @functools.wraps(func)
    def _log_function(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
            raise e
        else:
            logger.debug(
                f"Function {func.__name__} called",
                function=func.__name__,
                result=result,
                func_args=[str(arg) for arg in args],
                func_kwargs={k: str(v) for k, v in kwargs.items()},
            )
            return result

    return _log_function


Nothing = object()


def usecase(*funcs):
    """ Function composition with logging """
    funcs = list(funcs)

    def _usecase(
        data: typing.Any = Nothing, 
        executors_pool: typing.Optional[ExecutorsPool] = None,
        logger: typing.Optional[Logger] = default_logger,
    ):
        first_func = funcs[0]

        if logger: 
            first_func = log_function(first_func, logger)

        if data == Nothing:
            result = first_func()
        else:
            result = first_func(data)

        if executors_pool:
            result = executors_pool.execute(result)

        for func in funcs[1:]:
            if logger:
                func = log_function(func, logger)

            result = func(result)

            if executors_pool:
                result = executors_pool.execute(result)

        return result

    return _usecase
